/**
 * Copyright (c) 2007-2012 EBM WebSourcing, 2012-2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.se.eip.listener;

import static org.ow2.petals.se.eip.EIPConstants.ACTION;

import java.util.List;
import java.util.logging.Level;

import javax.jbi.messaging.MessagingException;
import javax.jbi.servicedesc.ServiceEndpoint;

import org.ow2.petals.commons.log.PetalsExecutionContext;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.jbidescriptor.generated.Consumes;
import org.ow2.petals.component.framework.listener.AbstractJBIListener;
import org.ow2.petals.component.framework.logger.Utils;
import org.ow2.petals.component.framework.process.async.AsyncContext;
import org.ow2.petals.component.framework.su.AbstractServiceUnitManager;
import org.ow2.petals.component.framework.su.ServiceUnitDataHandler;
import org.ow2.petals.se.eip.EIPComponent;
import org.ow2.petals.se.eip.ExchangeContext;
import org.ow2.petals.se.eip.async.CommonAsyncContext;
import org.ow2.petals.se.eip.patterns.Pattern;

/**
 * @author Adrien Louis - EBM WebSourcing
 */
public class JBIListener extends AbstractJBIListener implements ExchangeContext {

    /**
     * Return the Consumes elements of the service unit that contains the
     * provided endpoint
     * 
     * @param endpoint
     * @return a list of Consumes elements related to the provided endpoint
     */
    public List<Consumes> getSUConsumes(final ServiceEndpoint endpoint) {
        final AbstractServiceUnitManager sum = (AbstractServiceUnitManager) this.getComponent()
                .getServiceUnitManager();
        final ServiceUnitDataHandler sudh = sum.getSUDataHandlerForProvides(sum
                .getProvidesFromEndpoint(endpoint));
        return sudh.getDescriptor().getServices().getConsumes();
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.ow2.petals.component.framework.listener.AbstractJBIListener#
     * onAsyncJBIMessage
     * (org.ow2.petals.component.framework.api.message.Exchange,
     * org.ow2.petals.component.framework.process.async.AsyncContext)
     */
    @Override
    public boolean onAsyncJBIMessage(final Exchange asyncExchange, final AsyncContext asyncContext) {
        boolean response = true;
        if (asyncContext instanceof CommonAsyncContext) {
            response = this.process(asyncExchange, (CommonAsyncContext) asyncContext, false);
        } else {
            if (asyncExchange.isActiveStatus()) {
                asyncExchange.setError(new Exception("The context isn't recognized by the EIP"));
            }
        }
        return response;
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.ow2.petals.component.framework.listener.AbstractJBIListener#
     * onExpiredAsyncJBIMessage
     * (org.ow2.petals.component.framework.api.message.Exchange,
     * org.ow2.petals.component.framework.process.async.AsyncContext)
     */
    @Override
    public boolean onExpiredAsyncJBIMessage(final Exchange exchange, final AsyncContext asyncContext) {
        boolean response = true;
        if (asyncContext instanceof CommonAsyncContext) {
            response = this.process(exchange, (CommonAsyncContext) asyncContext, true);
        } else {
            if (exchange.isActiveStatus()) {
                exchange.setError(new Exception("The context isn't recognized by the EIP"));
            }
        }
        return response;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.listener.AbstractJBIListener#onJBIMessage
     * (org.ow2.petals.component.framework.api.message.Exchange)
     */
    @Override
    public boolean onJBIMessage(final Exchange exchange) {
        boolean result = true;

        // not interested of the ACKs
        if (exchange.isActiveStatus()) {
            if (exchange.isProviderRole()) {
                result = this.process(exchange, null, false);
            } else {
                this.processAsyncResponse(exchange);
            }
        }

        return result;
    }

    /**
     * Check for the action defined in the SU and call the related Pattern
     * processor
     * 
     * @param exchange
     */
    protected boolean process(final Exchange exchange, final CommonAsyncContext asyncContext,
            final boolean expired) {
        boolean result = true;
        String action = null;

        try {
            if (asyncContext == null) {
                if (this.getExtensions() != null) {
                    action = this.getExtensions().get(ACTION);
                }
            } else {
                action = asyncContext.getPattern();
            }
            if (action == null) {
                final String error = "action defined for the endpoint '"
                        + exchange.getEndpointName() + "' unknown by the EIP component";
                Utils.addMonitFailureTrace(this.getLogger(),
    					PetalsExecutionContext.getFlowAttributes(),
    					error, exchange.getRole());
                exchange.setError(new NoSuchMethodException(error));
            } else {
                final Pattern pattern = ((EIPComponent) this.getComponent()).getPatterns().get(
                        action.toLowerCase());
                if (pattern != null) {
                    if (asyncContext == null) {
                        result = pattern.processPattern(exchange, this);
                    } else {
                        result = pattern.processAsyncPattern(exchange, this, asyncContext, expired);
                    }
                } else {
                    final String error = "action '" + action + "' defined for the endpoint '"
                            + exchange.getEndpointName() + "' not supported by The EIP component.";
                    this.getLogger().severe(error);
                    Utils.addMonitFailureTrace(this.getLogger(),
        					PetalsExecutionContext.getFlowAttributes(),
        					error, exchange.getRole());
                    exchange.setError(new NoSuchMethodException(error));
                }
            }
        } catch (final MessagingException e) {
            this.getLogger().log(Level.SEVERE, e.getMessage(), e);
            Utils.addMonitFailureTrace(this.getLogger(),
					PetalsExecutionContext.getFlowAttributes(),
					e.getMessage(), exchange.getRole());
            exchange.setError(new Exception(e));
        }
        return result;
    }

    /**
     * Process a response from a previous asynchronous send performed by the EIP
     * component DONE and ERROR response are catch previously by the framework,
     * so this response is a Fault or an OUT message. Do nothing, the framework
     * will send back a DONE status
     */
    protected void processAsyncResponse(final Exchange exchange) {
        if (this.getLogger() != null && this.getLogger().isLoggable(Level.FINE)) {
            String itf = "";
            if (exchange.getEndpoint().getInterfaces().length != 0) {
                itf = exchange.getEndpoint().getInterfaces()[0].toString();
            }
            this.getLogger().fine(
                    "Receive an asynchronous OUT or Fault response which is not handled by an EIP pattern."
                            + "\nResponse received from ...\n - itf=" + itf + "\n - srv="
                            + exchange.getEndpoint().getServiceName() + "\n - ep ="
                            + exchange.getEndpoint().getEndpointName());
        }
    }
}
