/**
 * Copyright (c) 2007-2012 EBM WebSourcing, 2012-2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.se.eip.patterns;

import java.net.URI;
import java.util.logging.Logger;

import javax.jbi.messaging.Fault;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.xml.transform.Source;

import org.ow2.petals.commons.exception.ExceptionUtil;
import org.ow2.petals.component.framework.api.Message.MEPConstants;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.util.SourceUtil;
import org.ow2.petals.se.eip.EIPConstants;
import org.ow2.petals.se.eip.ExchangeContext;
import org.ow2.petals.se.eip.async.CommonAsyncContext;
import org.w3c.dom.Document;

/**
 * Provide a management of the source exchange response from the target
 * exchanges received. The responses must be aggregated into a same response.
 * 
 * @author Frederic Gardes - EBM WebSourcing
 */
public abstract class AbstractAggregatorPattern extends AbstractPattern {
    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.AbstractPattern#processAsync(org.ow2.petals
     * .component.framework.api.message.Exchange,
     * org.ow2.petals.se.eip.ExchangeContext,
     * org.ow2.petals.se.eip.async.CommonAsyncContext)
     */
    @Override
    public boolean processAsync(final Exchange exchange, final ExchangeContext context,
            final CommonAsyncContext asyncContext) {
        NormalizedMessage receivedMessage = null;
        boolean finished = false;

        if (this.logger == null) {
            this.logger = context.getLogger();
            if (this.logger == null) {
                this.logger = Logger.getLogger(AbstractPattern.class.getName());
            }
        }
        synchronized (asyncContext.getOriginalExchange()) {
            try {
                final int subMessages = asyncContext.decrementAndGetOngoingSubMessages();
                if (subMessages == 0) {
                    finished = true;
                }
                receivedMessage = PatternHelper.processSentExchange(exchange, context);
                if (!asyncContext.isTerminated()) {
                    if (receivedMessage != null) {
                        if (receivedMessage instanceof Fault) {
                            // fault
                            if (!asyncContext.isFaultRobust()) {
                                if (asyncContext.getOriginalExchange().getFault() == null
                                        && asyncContext.getOriginalExchange().getError() == null) {
                                    asyncContext.getOriginalExchange().setFault(
                                            (Fault) receivedMessage);
                                    finished = true;
                                }
                            } else {
                                // robustness on an external exchange
                                this.aggregate(asyncContext, receivedMessage);
                            }
                        } else {
                            // out
                            this.aggregate(asyncContext, receivedMessage);
                        }
                    } else {
                        // Done
                        Document document = this.getAggregatedDocument(asyncContext);
                        if (document == null) {
                            document = PatternHelper.createResultDocument(this.getNameSpace());
                        }
                        asyncContext
                                .getOriginalExchange()
                                .getOutMessage()
                                .setContent(
                                        SourceUtil.createStreamSource(document));
                    }
                }
            } catch (final MessagingException e) {
                // Error
                if (!asyncContext.isTerminated()) {
                    if (!asyncContext.isExceptionRobust()) {
                        asyncContext.getOriginalExchange().setError(e);
                        finished = true;
                    } else {
                        // robustness on an external exchange
                        try {
                            this.aggregate(asyncContext, e);
                        } catch (final MessagingException e2) {
                            asyncContext.getOriginalExchange().setError(e2);
                            finished = true;
                        } catch (final PEtALSCDKException e2) {
                            asyncContext.getOriginalExchange().setError(
                                    new MessagingException(ExceptionUtil.getExtendedMessage(e2)));
                            finished = true;
                        }
                    }
                }
            } catch (final PEtALSCDKException e) {
                // Error
                if (!asyncContext.isTerminated()) {
                    if (!asyncContext.isExceptionRobust()) {
                        asyncContext.getOriginalExchange().setError(
                                new MessagingException(ExceptionUtil.getExtendedMessage(e)));
                        finished = true;
                    } else {
                        // robustness on an external exchange
                        try {
                            this.aggregate(asyncContext, e);
                        } catch (final MessagingException e2) {
                            asyncContext.getOriginalExchange().setError(e2);
                            finished = true;
                        } catch (final PEtALSCDKException e2) {
                            asyncContext.getOriginalExchange().setError(
                                    new MessagingException(ExceptionUtil.getExtendedMessage(e2)));
                            finished = true;
                        }
                    }
                }
            }
            if (finished) {
                this.logger.fine(this.getName() + ": terminate the pattern");
                if (!asyncContext.isTerminated()) {
                    try {
                        if (asyncContext.getOriginalExchange().getFault() == null
                                && asyncContext.getOriginalExchange().getError() == null) {
                            final Document document = SourceUtil
                                    .createDocument(
                                            asyncContext.getOriginalExchange().getOutMessage()
                                                    .getContent());
                            document.normalizeDocument();
                            asyncContext
                                    .getOriginalExchange()
                                    .getOutMessage()
                                    .setContent(
                                            SourceUtil
                                                    .createStreamSource(document));
                        }
                        context.send(asyncContext.getOriginalExchange());
                    } catch (final MessagingException e) {
                        this.logger.warning(this.getName() + ": unable to send an exchange, "
                                + e.getMessage());
                    } catch (final PEtALSCDKException e) {
                        this.logger.warning(this.getName() + ": unable to send an exchange, "
                                + e.getMessage());
                    } finally {
                        asyncContext.setTerminated();
                    }
                }
            }
        }
        return false;
    }

    /**
     * Aggregate the string content of the exception to the source exchange
     * 
     * @param asyncContext
     *            The context where the source exchange is stored
     * @param exception
     *            The received exception
     * @throws MessagingException
     * @throws PEtALSCDKException
     */
    private void aggregate(final CommonAsyncContext asyncContext, final Exception exception)
            throws MessagingException, PEtALSCDKException {
        // get the document from the source exchange
        Document aggregatedDocument = this.getAggregatedDocument(asyncContext);
        final Document itemDocument = PatternHelper.createExceptionDocument(this.getNameSpace());
        if (itemDocument != null) {
            itemDocument.getDocumentElement().setTextContent(exception.getMessage());
            aggregatedDocument = PatternHelper.aggregateContent(itemDocument, aggregatedDocument,
                    EIPConstants.RESULT_ELEMENT, this.getNameSpace());
            asyncContext.getOriginalExchange().getOutMessage()
                    .setContent(SourceUtil.createStreamSource(aggregatedDocument));
        }
    }

    /**
     * Aggregate the XML content, attachments and properties at the content of
     * the out normalized message to the source exchange
     * 
     * @param asyncContext
     *            The context where the source exchange is stored
     * @param receivedMessage
     *            The received normalized message
     * @throws MessagingException
     * @throws PEtALSCDKException
     */
    private void aggregate(final CommonAsyncContext asyncContext,
            final NormalizedMessage receivedMessage) throws MessagingException, PEtALSCDKException {
        // get the document from the source exchange
        Document aggregatedDocument = this.getAggregatedDocument(asyncContext);
        aggregatedDocument = PatternHelper.aggregateContent(receivedMessage, aggregatedDocument,
                EIPConstants.RESULT_ELEMENT, this.getNameSpace());
        final Source source = SourceUtil.createStreamSource(aggregatedDocument);
        asyncContext.getOriginalExchange().getOutMessage().setContent(source);
        PatternHelper.copyAttachments(receivedMessage, asyncContext.getOriginalExchange()
                .getOutMessage());
        PatternHelper.copyProperties(receivedMessage, asyncContext.getOriginalExchange()
                .getOutMessage());
    }

    /**
     * Get the aggregated document into the source exchange
     * 
     * @param asyncContext
     *            The context with the source exchange
     * @return The document build from the content of the out normalized message
     * @throws PEtALSCDKException
     */
    private Document getAggregatedDocument(final CommonAsyncContext asyncContext)
            throws PEtALSCDKException {
        Document aggregatedDocument = null;
        this.logger.fine(this.getName() + ": aggregates a document to the original out message");
        final NormalizedMessage normalizedMessage = asyncContext.getOriginalExchange()
                .getOutMessage();
        if (normalizedMessage != null) {
            final Source outContent = normalizedMessage.getContent();
            if (outContent != null) {
                aggregatedDocument = SourceUtil.createDocument(outContent);
            }
        }
        return aggregatedDocument;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.AbstractPattern#validateMEP(java.net.URI)
     */
    @Override
    public boolean validateMEP(final URI mep) {
        return MEPConstants.IN_OUT_PATTERN.equals(mep);
    }
}
