/**
 * Copyright (c) 2007-2012 EBM WebSourcing, 2012-2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.se.eip.patterns;

import java.net.URI;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.messaging.MessagingException;

import org.ow2.petals.commons.exception.ExceptionUtil;
import org.ow2.petals.component.framework.api.configuration.ConfigurationExtensions;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.jbidescriptor.generated.Consumes;
import org.ow2.petals.se.eip.EIPConstants;
import org.ow2.petals.se.eip.ExchangeContext;
import org.ow2.petals.se.eip.async.CommonAsyncContext;

/**
 * @author Christophe HAMERLING - EBM WebSourcing
 */
public abstract class AbstractPattern implements Pattern {

    protected Logger logger;

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.Pattern#processPattern(org.ow2.petals.
     * component.framework.util.Exchange, org.ow2.petals.se.eip.ExchangeContext)
     */
    public final boolean processPattern(final Exchange exchange, final ExchangeContext context) {
        boolean result = true;
        if (this.logger == null) {
            this.logger = context.getLogger();
            if (this.logger == null) {
                this.logger = Logger.getLogger(AbstractPattern.class.getName());
            }
        }
        if (this.logger.isLoggable(Level.FINE)) {
            this.logger.fine(this.getName() + ": processing the pattern");
        }
        if (this.validateMEPPattern(exchange) && exchange.isActiveStatus()) {
            result = this.process(exchange, context);
        }
        return result;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.Pattern#processAsyncPattern(org.ow2.petals
     * .component.framework.api.message.Exchange,
     * org.ow2.petals.se.eip.ExchangeContext,
     * org.ow2.petals.se.eip.async.CommonAsyncContext, boolean)
     */
    public final boolean processAsyncPattern(final Exchange exchange,
            final ExchangeContext context, final CommonAsyncContext asyncContext,
            final boolean expired) {
        if (this.logger == null) {
            this.logger = context.getLogger();
            if (this.logger == null) {
                this.logger = Logger.getLogger(AbstractPattern.class.getName());
            }
        }
        if (this.logger.isLoggable(Level.FINE)) {
            this.logger.fine(this.getName() + ": processing asynchronously the pattern");
        }
        try {
            if (expired) {
                if (this.logger.isLoggable(Level.FINE)) {
                    this.logger.fine(this.getName() + ": the process is expired");
                }
                this.processExpiredAsync(exchange, context, asyncContext);
            } else {
                this.processAsync(exchange, context, asyncContext);
            }
        } catch (final MessagingException e) {
            if (this.logger.isLoggable(Level.FINE)) {
                this.logger.fine(this.getName() + ": failed to process the async pattern, "
                        + e.getMessage());
            }
            if (exchange.isActiveStatus()) {
                exchange.setError(e);
            }
        } catch (final PEtALSCDKException e) {
            if (this.logger.isLoggable(Level.FINE)) {
                this.logger.fine(this.getName() + ": failed to process the async pattern, "
                        + e.getMessage());
            }
            if (exchange.isActiveStatus()) {
                exchange.setError(new MessagingException(ExceptionUtil.getExtendedMessage(e)));
            }
        }
        return false;
    }

    /**
     * Allow to log an exchange sent to a <Consumes>
     * @param consumedService
     * @param exchange
     */
    protected final void logSend(final Consumes consumedService, final Exchange exchange) {
        if (this.logger.isLoggable(Level.FINE)) {
            String interfaceName = null;
            String serviceName = null;
            if (consumedService.getInterfaceName() != null) {
                interfaceName = consumedService.getInterfaceName().toString();
            }
            if (consumedService.getServiceName() != null) {
                serviceName = consumedService.getServiceName().toString();
            }

            this.logger.fine(this.getName() + ": invocation of the service : interface name="
                    + interfaceName + " / service name=" + serviceName + " / endpoint name="
                    + consumedService.getEndpointName());
        }
    }

    /**
     * Process the message.
     * 
     * @param exchange
     * @param context
     */
    public abstract boolean process(Exchange exchange, ExchangeContext context);

    /**
     * Process the async message.
     * 
     * @param exchange
     * @param context
     */
    public abstract boolean processAsync(final Exchange exchange, final ExchangeContext context,
            final CommonAsyncContext asyncContext);

    /**
     * Process the expired async message.
     * 
     * @param exchange
     * @param context
     */
    public boolean processExpiredAsync(final Exchange exchange, final ExchangeContext context,
            final CommonAsyncContext asyncContext) throws MessagingException, PEtALSCDKException {
        exchange.setError(new MessagingException(
                "Failed to process the asynchronous pattern, timeout reached!"));
        return true;
    }

    /**
     * Validate the MEP for the pattern. This method is called on each message
     * exchange in order to validate that the MEP is compatible.
     * 
     * @param mep
     *            the mep of the incoming message.
     * @return true if the mep is compatible, false otherwise.
     */
    public abstract boolean validateMEP(URI mep);

    /**
     * get the fault robustness into the extensions
     * 
     * @param extensions
     *            the ConfigurationExtensions used
     * @return true if the pattern have to be fault save, false else, and false
     *         if not found
     */
    protected boolean isFaultRobust(final ConfigurationExtensions extensions) {
        for (final String param : extensions.keySet()) {
            if (EIPConstants.FAULT_ROBUST.equals(param)) {
                return extensions.get(param).equals("true");
            }
        }
        return false;
    }

    /**
     * get the exception robustness into the extensions
     * 
     * @param extensions
     *            the ConfigurationExtensions used
     * @return true if the pattern have to be exception save, false else, and
     *         false if not found
     */
    protected boolean isExceptionRobust(final ConfigurationExtensions extensions) {
        for (final String param : extensions.keySet()) {
            if (EIPConstants.EXCEPTION_ROBUST.equals(param)) {
                return extensions.get(param).equals("true");
            }
        }
        return false;
    }

    /**
     * get the fault to exception parameter into the extensions
     * 
     * @param extensions
     *            the ConfigurationExtensions used
     * @return true if the pattern have to transform the faults to exceptions,
     *         false else, and false if not found
     */
    protected boolean isFaultToException(final ConfigurationExtensions extensions) {
        for (final String param : extensions.keySet()) {
            if (EIPConstants.FAULT_TO_EXCEPTION.equals(param)) {
                return extensions.get(param).equals("true");
            }
        }
        return false;
    }

    /**
     * get the attachment mode into the extensions
     * 
     * @param extensions
     *            the ConfigurationExtensions used
     * @return true if the pattern have to treat the attachments, false else,
     *         and false if not found
     */
    protected boolean isAttachmentMode(final ConfigurationExtensions extensions) {
        for (final String param : extensions.keySet()) {
            if (EIPConstants.ATTACHMENT_MODE.equals(param)) {
                return extensions.get(param).equals("true");
            }
        }
        return false;
    }

    /**
     * Validate the MEP pattern of the exchange
     * 
     * @param exchange
     * @return
     */
    private final boolean validateMEPPattern(final Exchange exchange) {
        boolean result = true;

        URI pattern = null;
        try {
            pattern = exchange.getExchangePattern();
        } catch (final MessagingException e) {
            exchange.setError(e);
        }

        if (!this.validateMEP(pattern)) {
            exchange.setError(new MessagingException("The " + this.getName()
                    + " pattern does not support " + pattern + " pattern"));
            result = false;
        }
        return result;
    }

    /**
     * Get the pattern namespace.
     * 
     * @return The namespace
     */
    public abstract String getNameSpace();
}
