/**
 * Copyright (c) 2007-2012 EBM WebSourcing, 2012-2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.se.eip.patterns;

import java.net.URI;
import java.util.List;
import java.util.logging.Level;

import javax.jbi.messaging.MessagingException;

import org.ow2.petals.commons.exception.ExceptionUtil;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.jbidescriptor.generated.Consumes;
import org.ow2.petals.component.framework.util.ExchangeUtil;
import org.ow2.petals.se.eip.EIPConstants;
import org.ow2.petals.se.eip.ExchangeContext;
import org.ow2.petals.se.eip.async.CommonAsyncContext;

/**
 * Change the MEP of an exchange. 
 * <p>
 * Use the MEP defined in the Consumes element. <br>
 * Depending on the MEP, some response can be lost. No robustness on fault or
 * exception: this concept is linked to at least 2 target exchanges, 2 consumes
 * </p>
 * 
 * @author Adrien Louis - EBM WebSourcing
 * @author Christophe Hamerling - EBM WebSourcing
 */
public class Bridge extends AbstractMEPBridgeWithUnicConsumesPattern {

    public final static String BRIDGE_NAMESPACE = EIPConstants.NAMESPACE + "/bridge";

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.Pattern#processPattern(org.ow2.petals.
     * component.framework.util.Exchange, org.ow2.petals.se.eip.ExchangeContext)
     */
    @Override
    public boolean process(final Exchange exchange, final ExchangeContext context) {
        try {
            final List<Consumes> consumesList = context.getSUConsumes(exchange.getEndpoint());
            if (consumesList != null && consumesList.size() == 1) {
                final Consumes consumes = consumesList.get(0);
                // create the exchange with the new MEP to send to the next
                // service
                final Exchange newExchange = context.createConsumeExchange(consumes);
                if (newExchange.getOperation() == null) {
                    newExchange.setOperation(exchange.getOperation());
                }
                PatternHelper.copy(exchange.getInMessage(), newExchange.getInMessage());
                ExchangeUtil.copyExchangeProperties(exchange, newExchange);
                // send the exchange to the next service
                if (this.logger.isLoggable(Level.FINE)) {
                    this.logger.fine(this.getName() + ": the source MEP is "
                            + exchange.getPattern());
                    this.logger.fine(this.getName() + ": the target MEP is "
                            + newExchange.getPattern());
                }
                this.logSend(consumes, newExchange);
                final boolean faultRobust = this.isFaultToException(context.getExtensions());
                final CommonAsyncContext asyncContext = new CommonAsyncContext(exchange,
                        CommonAsyncContext.DEFAULT_TIME_TO_LIVE, this.getName(), 1, false, false,
                        faultRobust);
                context.sendAsync(newExchange, asyncContext);
            } else {
                exchange.setError(new MessagingException(this.getName()
                        + ": the Service Unit must define 1 Consumes section"));
            }
        } catch (final MessagingException e) {
            exchange.setError(e);
        } catch (final PEtALSCDKException e) {
            exchange.setError(new MessagingException(ExceptionUtil.getExtendedMessage(e)));
        }

        return false;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.Pattern#init()
     */
    public void init() {
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.Pattern#getName()
     */
    public String getName() {
        return "Bridge";
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.AbstractPattern#validateMEP(java.net.URI)
     */
    @Override
    public boolean validateMEP(final URI mep) {
        return true;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.AbstractPattern#getNameSpace()
     */
    @Override
    public String getNameSpace() {
        return BRIDGE_NAMESPACE;
    }
}
