/**
 * Copyright (c) 2007-2012 EBM WebSourcing, 2012-2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.se.eip.patterns;

import java.io.InputStream;
import java.net.URI;
import java.util.List;
import java.util.logging.Level;

import javax.jbi.JBIException;
import javax.jbi.messaging.Fault;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.xml.transform.stream.StreamSource;

import org.ow2.petals.commons.exception.ExceptionUtil;
import org.ow2.petals.component.framework.api.Message.MEPConstants;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.jbidescriptor.generated.Consumes;
import org.ow2.petals.component.framework.util.ExchangeUtil;
import org.ow2.petals.component.framework.util.SourceUtil;
import org.ow2.petals.se.eip.EIPConstants;
import org.ow2.petals.se.eip.ExchangeContext;
import org.ow2.petals.se.eip.async.CommonAsyncContext;

/**
 * Dispatch the exchange to all Consumes services defined in the service unit.
 * <p>Accept InOnly.</p>
 * 
 * @author Adrien Louis - EBM WebSourcing
 * @author Christophe Hamerling - EBM WebSourcing
 * @author Frederic Gardes - EBM WebSourcing
 */
public class Dispatcher extends AbstractMEPBridgePattern {

    public final static String DISPATCHER_NAMESPACE = EIPConstants.NAMESPACE + "/dispatcher";

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.Pattern#processPattern(org.ow2.petals.
     * component.framework.util.Exchange, org.ow2.petals.se.eip.ExchangeContext)
     */
    @Override
    public boolean process(final Exchange exchange, final ExchangeContext context) {
        try {
            final List<Consumes> consumesList = context.getSUConsumes(exchange.getEndpoint());
            if (consumesList != null && consumesList.size() != 0) {
                this.logger.fine("Begin asynchronous exchanges send to consumes for exchange["+exchange.getExchangeId()+"]");
                this.sendAsyncExchanges(consumesList, exchange, context);
            } else {
                this.logger.warning("No consume section was define for service unit");
                exchange.setError(new MessagingException(
                        this.getName()
                                + ": the Service Unit must define at least 1 Consumes sections for the pattern "));
            }
        } catch (final PEtALSCDKException e) {
            this.logger.log(Level.WARNING, "PEtALSCDKException occurred during Dispatcher processing for ["+exchange.getExchangeId()+"] exchange", e);
            exchange.setError(new MessagingException(ExceptionUtil.getExtendedMessage(e)));
        } catch (final MessagingException e) {
            this.logger.log(Level.WARNING, "MessagingException occurred during Dispatcher processing for ["+exchange.getExchangeId()+"] exchange", e);
            exchange.setError(e);
        }

        return false;
    }

    /**
     * Check InOnly or RobustInOnly pattern. <br>
     * 
     * @param uri
     * @return true if MEP is InOnly, false otherwise
     */
    @Override
    public boolean validateMEP(final URI uri) {
        return MEPConstants.IN_ONLY_PATTERN.equals(uri);
    }

    /**
     * Send a copy of the exchange's IN message to all the services to consume.
     * Consumes represent 'performers entities' in Dispatcher Pattern
     * @param consumesList
     * @param exchange
     * @param context
     * @throws JBIException
     * @throws PEtALSCDKException
     */
    private void sendAsyncExchanges(final List<Consumes> consumesList, final Exchange exchange,
            final ExchangeContext context) throws PEtALSCDKException, MessagingException {

        // send all exchanges asynchronously
        final boolean faultToException = this.isFaultToException(context.getExtensions());
        final CommonAsyncContext asyncContext = new CommonAsyncContext(exchange,
                CommonAsyncContext.DEFAULT_TIME_TO_LIVE, this.getName(), consumesList.size(),
                false, false, faultToException);

        final NormalizedMessage inMessage = exchange.getInMessage();
        InputStream inMessageAsStream = PatternHelper.handleStreamSource(inMessage);
        this.logger.fine("Begin "+this.getName()+" sendAsyncExchange() for exchange["+exchange.getExchangeId()+"] with FaultToException policy: "+faultToException
                +" and default time to live: "+CommonAsyncContext.DEFAULT_TIME_TO_LIVE);
        /*
         * Dispatch for each consumers (performers) a copy
         * of incoming IN message and copy its content,
         * properties, ...
         */
        for (final Consumes toConsume : consumesList) {
            final Exchange dipatchedExchange = context.createConsumeExchange(toConsume);
            if (dipatchedExchange.getOperation() == null) {
                dipatchedExchange.setOperation(exchange.getOperation());
            }
            PatternHelper.copy(inMessage, dipatchedExchange.getInMessage());
            ExchangeUtil.copyExchangeProperties(exchange, dipatchedExchange);
            this.logSend(toConsume, dipatchedExchange);
            context.sendAsync(dipatchedExchange, asyncContext);

            // Recopy the forwarded StreamSource
            if (inMessageAsStream != null) {
                // reset the Stream source for next iteration
                inMessage.setContent(new StreamSource(inMessageAsStream));
                inMessageAsStream = PatternHelper.handleStreamSource(inMessage);
            }
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.Pattern#init()
     */
    public void init() {
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.Pattern#getName()
     */
    public String getName() {
        return "Dispatcher";
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.AbstractPattern#processAsync(org.ow2.petals
     * .component.framework.api.message.Exchange,
     * org.ow2.petals.se.eip.ExchangeContext,
     * org.ow2.petals.se.eip.async.CommonAsyncContext)
     */
    @Override
    public boolean processAsync(final Exchange exchange, final ExchangeContext context,
            final CommonAsyncContext asyncContext) {

        try {
            final int subMessages = asyncContext.decrementAndGetOngoingSubMessages();
            this.logger.fine(subMessages+" sub messages wait for process async for exchange["+exchange.getExchangeId()+"]");
            if (subMessages == 0) {
                this.logger.fine("No sub message wait for process async");
                this.processResponse(exchange, context, asyncContext);
            } else {
                // process the step response
                final NormalizedMessage receivedMessage = PatternHelper.processSentExchange(
                        exchange, context);
                if (!asyncContext.isTerminated() && receivedMessage != null
                        && receivedMessage instanceof Fault) {
                    this.logger.fine("Received message is a Fault and is not terminated");
                    if (asyncContext.isFaultToException()) {
                        asyncContext.getOriginalExchange().setError(
                                new MessagingException(SourceUtil.createString(
                                        receivedMessage.getContent())));

                        this.logger.fine("The asynchrone context handle a FaultToException and return to" +
                        		" the original exchange the following MessageException: "+asyncContext.getOriginalExchange().getError().getMessage());
                        // terminate the original exchange
                        context.send(asyncContext.getOriginalExchange());
                        asyncContext.setTerminated();
                    }
                }
            }
        } catch (final MessagingException e) {
                    
            if (!asyncContext.isTerminated()) {
                this.logger.warning("Messaging exception was catched: "+e.getMessage()+" and " +
                        "the async context isn't terminated, so the original exchange must" +
                        " be terminate by an error");
                asyncContext.getOriginalExchange().setError(e);
                // in this case, it's an error, terminate the original exchange
                try {
                    context.send(asyncContext.getOriginalExchange());
                } catch (final MessagingException e1) {
                    this.logger.warning(this.getName()
                            + ": unable to close the source exchange on error status");
                } finally {
                    asyncContext.setTerminated();
                }
            }
        } catch (PEtALSCDKException e) {
            if (!asyncContext.isTerminated()) {
                this.logger.warning("PEtALSCDK exception was catched: "+e.getMessage()+" and " +
                        "the async context isn't terminated, so the original exchange must" +
                        " be terminate by an error");
                asyncContext.getOriginalExchange().setError(
                        new MessagingException(ExceptionUtil.getExtendedMessage(e)));
                // in this case, it's an error, terminate the original exchange
                try {
                    context.send(asyncContext.getOriginalExchange());
                } catch (final MessagingException e1) {
                    this.logger.warning(this.getName()
                            + ": unable to close the source exchange on error status");
                } finally {
                    asyncContext.setTerminated();
                }
            }
        }

        return true;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.AbstractPattern#getNameSpace()
     */
    @Override
    public String getNameSpace() {
        return DISPATCHER_NAMESPACE;
    }
}
