/**
 * Copyright (c) 2007-2012 EBM WebSourcing, 2012-2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.se.eip.patterns;

import java.net.URI;
import java.util.List;

import javax.jbi.messaging.MessagingException;
import javax.xml.xpath.XPathExpressionException;

import org.ow2.petals.commons.exception.ExceptionUtil;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.jbidescriptor.generated.Consumes;
import org.ow2.petals.component.framework.util.ExchangeUtil;
import org.ow2.petals.se.eip.EIPConstants;
import org.ow2.petals.se.eip.ExchangeContext;
import org.ow2.petals.se.eip.async.CommonAsyncContext;

/**
 * A Router pattern on the content of the In message. 
 * <p>
 * No robustness on fault or exception: this concept is linked to at least 2 target exchanges, 2 consumes.
 * </p>
 * 
 * @author Adrien Louis - EBM WebSourcing
 * @author Olivier Fabre - EBM WebSourcing
 * @author Christophe Hamerling - EBM WebSourcing
 * @author Frederic Gardes - EBM WebSourcing
 */
public class Router extends AbstractMEPBridgeWithUnicConsumesPattern {

    public final static String ROUTER_NAMESPACE = EIPConstants.NAMESPACE + "/router";

    private XPathRouterEngine xpathRouterEngine = new XPathRouterEngine();

    private OperationRouterEngine operationRouterEngine = new OperationRouterEngine();

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.AbstractPattern#process(org.ow2.petals
     * .component.framework.util.Exchange,
     * org.ow2.petals.se.eip.ExchangeContext)
     */
    @Override
    public boolean process(final Exchange exchange, final ExchangeContext context) {
        final RouterEngine routerEngine;
        final List<Consumes> consumesList = context.getSUConsumes(exchange.getEndpoint());
        List<String> conditions = this.operationRouterEngine.getOperationParameters(context
                .getExtensions());
        if (conditions.isEmpty()) {
            conditions = this.xpathRouterEngine.getXpathParameters(context.getExtensions());
            routerEngine = this.xpathRouterEngine;
        } else {
            routerEngine = this.operationRouterEngine;
        }

        if (conditions.isEmpty()) {
            exchange.setError(new MessagingException(this.getName()
                    + ": the Service Unit must at least one routing parameter"));
        }

        try {
            // there must be numberOfConditions+1 consumes defined in the SU
            if (consumesList != null) {
                if (consumesList.size() == conditions.size() + 1) {
                    // route to get the target Consumes
                    final int consumesIndex = routerEngine.route(conditions, exchange);

                    // forward the exchange to the correct service, depending on
                    // the test result
                    final Consumes consumedService = consumesList.get(consumesIndex);
                    final Exchange routedExchange = context.createConsumeExchange(consumedService);
                    if (routedExchange.getOperation() == null) {
                        routedExchange.setOperation(exchange.getOperation());
                    }
                    PatternHelper.copy(exchange.getInMessage(), routedExchange.getInMessage());
                    ExchangeUtil.copyExchangeProperties(exchange, routedExchange);

                    this.logSend(consumedService, routedExchange);

                    final boolean faultRobust = this.isFaultToException(context.getExtensions());
                    final CommonAsyncContext asyncContext = new CommonAsyncContext(exchange,
                            CommonAsyncContext.DEFAULT_TIME_TO_LIVE, this.getName(), 1, false,
                            false, faultRobust);

                    context.sendAsync(routedExchange, asyncContext);
                } else {
                    exchange.setError(new MessagingException(
                            this.getName()
                                    + ": the Service Unit must define (numberOfRoutingParameters+1) Consumes sections"));
                }
            } else {
                exchange.setError(new MessagingException(
                        this.getName()
                                + ": the Service Unit must define (numberOfRoutingParameters+2) Consumes sections"));
            }

        } catch (final MessagingException e) {
            exchange.setError(e);
        } catch (final PEtALSCDKException e) {
            exchange.setError(new MessagingException(ExceptionUtil.getExtendedMessage(e)));
        } catch (final XPathExpressionException e) {
            exchange.setError(e);
        } catch (final IllegalArgumentException e) {
            // this exception is thrown by the Qname.valueof() method of the
            // OperationRouterEngine
            exchange.setError(e);
        }

        return false;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.Pattern#getName()
     */
    public String getName() {
        return "Router";
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.AbstractPattern#validateMEP(java.net.URI)
     */
    @Override
    public boolean validateMEP(final URI mep) {
        return true;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.Pattern#init()
     */
    public void init() {
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.AbstractPattern#getNameSpace()
     */
    @Override
    public String getNameSpace() {
        return ROUTER_NAMESPACE;
    }
}
