/**
 * Copyright (c) 2007-2012 EBM WebSourcing, 2012-2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.se.eip.patterns.mock;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.URI;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.Fault;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;

import org.ow2.petals.commons.log.FlowAttributes;
import org.ow2.petals.jbi.messaging.exchange.MessageExchangeImpl;

/**
 * @author Frederic Gardes - EBM WebSourcing
 */
public class MessageExchangeMock implements MessageExchange {

    public static final String IN_MSG = "in";

    public static final URI IN_ONLY_PATTERN = URI.create("http://www.w3.org/2004/08/wsdl/in-only");

    public static final URI IN_OPTIONAL_OUT_PATTERN = URI
    .create("http://www.w3.org/2004/08/wsdl/in-opt-out");

    public static final URI IN_OUT_PATTERN = URI.create("http://www.w3.org/2004/08/wsdl/in-out");

    public static final String OUT_MSG = "out";

    public static final URI ROBUST_IN_ONLY_PATTERN = URI
    .create("http://www.w3.org/2004/08/wsdl/robust-in-only");

    private static final short SERIALIZE_ROLE_CONSUMER = 0;

    private static final short SERIALIZE_ROLE_PROVIDER = 1;

    private static final short SERIALIZE_STATUS_ACTIVE = 0;

    private static final short SERIALIZE_STATUS_DONE = 1;

    private static final short SERIALIZE_STATUS_ERROR = 2;

    protected ServiceEndpoint consumerEndpoint;

    protected ServiceEndpoint endpoint;

    protected Exception error;

    protected String exchangeId;
    
    protected FlowAttributes flowAttributes;

    protected Fault fault;

    // //////////////////////////////////////////////////////////
    //
    // implementation specific attributes
    //
    // //////////////////////////////////////////////////////////

    protected QName interfaceName;

    protected Map<String, NormalizedMessage> messages = new HashMap<String, NormalizedMessage>();

    protected QName operation;

    protected URI pattern;

    protected Map<String, Object> properties = new HashMap<String, Object>();

    protected transient Role role;

    protected QName service;

    protected transient ExchangeStatus status;

    protected boolean terminated;

    protected boolean transacted;

    private boolean checkRole = false;

    /**
     * Create a MessageExchangeImpl This constructor is called by a consumer who
     * wants to initialize an exchange.
     * 
     * @param consumerEndpoint
     * @param uri
     * @param role
     * @param status
     */
    public MessageExchangeMock(final ServiceEndpoint consumerEndpoint, final URI uri,
            final Role role, final ExchangeStatus status) {
        this.status = status;
        this.role = role;
        this.consumerEndpoint = consumerEndpoint;
        this.pattern = uri;
        this.checkRole = true;
    }

    /**
     * Create a MessageExchangeImpl This constructor is called by a consumer who
     * wants to initialize an exchange.
     * 
     * @param uri
     * 
     * @param consumerDeliveryChannel
     */
    public MessageExchangeMock(final URI uri, final Role role) {
        this.status = ExchangeStatus.ACTIVE;
        this.role = role;
        this.pattern = uri;
        this.checkRole = true;
    }

    /**
     * Set the checkRole
     * 
     * @param checkRole
     *            the checkRole to set
     */
    public void setCheckRole(final boolean checkRole) {
        this.checkRole = checkRole;
    }

    /**
     * @see javax.jbi.messaging.MessageExchange#createFault()
     */
    public Fault createFault() throws MessagingException {
        return new FaultMock();
    }

    /**
     * @see javax.jbi.messaging.MessageExchange#createMessage()
     */
    public NormalizedMessage createMessage() throws MessagingException {
        return new NormalizedMessageMock();
    }

    public ServiceEndpoint getConsumerEndpoint() {
        return this.consumerEndpoint;
    }

    /**
     * @see javax.jbi.messaging.MessageExchange#getEndpoint()
     */
    public ServiceEndpoint getEndpoint() {
        return this.endpoint;
    }

    /**
     * {@inheritDoc}
     */
    public ServiceEndpoint getServiceEndpoint() {
        return this.endpoint;
    }

    /**
     * @see javax.jbi.messaging.MessageExchange#getError()
     */
    public Exception getError() {
        return this.error;
    }

    /**
     * @see javax.jbi.messaging.MessageExchange#getExchangeId()
     */
    public String getExchangeId() {
        return this.exchangeId;
    }

    /**
     * @see javax.jbi.messaging.MessageExchange#getFault()
     */
    public Fault getFault() {
        return this.fault;
    }

    public QName getInterfaceName() {
        return this.interfaceName;
    }

    /**
     * get the message associated to the reference. It is <b>Not</b> Case
     * Sensitive.
     * 
     * @see javax.jbi.messaging.MessageExchange#getMessage(java.lang.String)
     */
    public NormalizedMessage getMessage(final String name) {
        if (name != null) {
            return this.messages.get(name.toLowerCase());
        } else {
            return null;
        }
    }

    public Map<String, NormalizedMessage> getMessages() {
        return this.messages;
    }

    /**
     * @see javax.jbi.messaging.MessageExchange#getOperation()
     */
    public QName getOperation() {
        return this.operation;
    }

    /**
     * @see javax.jbi.messaging.MessageExchange#getPattern()
     */
    public URI getPattern() {
        return this.pattern;
    }

    /**
     * @see javax.jbi.messaging.MessageExchange#getProperty(java.lang.String)
     */
    public Object getProperty(final String name) {
        return this.properties.get(name);
    }

    /**
     * @see javax.jbi.messaging.MessageExchange#getPropertyNames()
     */
    public Set<?> getPropertyNames() {
        return this.properties.keySet();
    }

    /**
     * (non-Javadoc)
     * 
     * @see javax.jbi.messaging.MessageExchange#getRole()
     */
    public Role getRole() {
        return this.role;
    }

    /**
     * @see javax.jbi.messaging.MessageExchange#getService()
     */
    public QName getService() {
        return this.service;
    }

    /**
     * @see javax.jbi.messaging.MessageExchange#getStatus()
     */
    public javax.jbi.messaging.ExchangeStatus getStatus() {
        return this.status;
    }

    public boolean isTerminated() {
        return this.terminated;
    }

    /**
     * @see javax.jbi.messaging.MessageExchange#isTransacted()
     */
    public boolean isTransacted() {
        return this.transacted;
    }

    /**
     * {@inheritDoc}
     */
    public void setEndpoint(final ServiceEndpoint endpoint) {
        this.endpoint = endpoint;
    }

    /**
     * @see javax.jbi.messaging.MessageExchange#setError(java.lang.Exception)
     */
    public void setError(final Exception error) {
        this.error = error;
        try {
            this.setStatus(ExchangeStatus.ERROR);
        } catch (final MessagingException e) {
            // the status can not be changed at this transition
        }
    }

    /**
     * @param exchangeId
     *            The exchangeId to set.
     */
    public void setExchangeId(final String exchangeId) {
        this.exchangeId = exchangeId;
    }

    // //////////////////////////////////////////
    // specific methods
    // //////////////////////////////////////////

    /**
     * (non-Javadoc)
     * 
     * @see javax.jbi.messaging.MessageExchange#setFault(javax.jbi.messaging.Fault)
     */
    public void setFault(final Fault fault) throws MessagingException {
        this.checkNotTerminated();

        if (this.role == null) {
            throw new MessagingException("The Role is not defined.");
        }

        // check that the status has not be set with DONE or EXCEPTION
        if (ExchangeStatus.DONE.equals(this.status) || ExchangeStatus.ERROR.equals(this.status)) {
            throw new MessagingException(
            "The MessageExchange state and the current Role do not allow this operation.");
        }
        if (this.checkRole) {
            // only the provider can send a fault, except in the case of a
            // InOptionnalOut,
            // where the consumer can set a fault after the provider sent a
            // response

            if (MessageExchange.Role.CONSUMER.equals(this.role)
                    && !(IN_OPTIONAL_OUT_PATTERN.equals(this.pattern) && this.messages
                            .containsKey(OUT_MSG))) {
                throw new MessagingException(
                "The MessageExchange state and the current Role do not allow this operation.");
            }
        }

        this.fault = fault;
    }

    public void setInterfaceName(final QName interfaceName) {
        this.interfaceName = interfaceName;
    }

    /**
     * Set the specified message. Check the availability to set the message with
     * the following rules: <br>
     * <li>CONSUMER Role <-> in reference or PROVIDER Role <-> out reference</li>
     * <li>JBI predefined MEP <-> in or out reference</li> <li>setMessage(XXX
     * ,reference) has not already been called.<br>
     * This is <b>not</b> case sensitive
     * 
     * @see javax.jbi.messaging.MessageExchange#setMessage(javax.jbi.messaging.
     *      NormalizedMessage, java.lang.String)
     */
    public void setMessage(final NormalizedMessage msg, final String name)
    throws MessagingException {
        this.checkNotTerminated();

        if (msg == null) {
            throw new MessagingException("NormalizedMessage must be non null.");
        }
        if (name == null || name.trim().length() == 0) {
            throw new MessagingException("The message reference must be non null and non empty.");
        }

        // check if the pattern allows this reference
        this.checkPatternMatching(name);

        // check that the message has not already been set
        if (this.messages.containsKey(name.toLowerCase())) {
            throw new MessagingException("A message has already been set with the '" + name
                    + "' reference.");
        }

        // no exception was thrown, set the message
        this.messages.put(name.toLowerCase(), msg);
    }

    /**
     * @see javax.jbi.messaging.MessageExchange#setOperation(javax.xml.namespace.
     *      QName)
     */
    public void setOperation(final QName name) {
        this.operation = name;
    }

    /**
     * @param pattern
     *            The pattern to set.
     */
    public void setPattern(final URI pattern) {
        this.pattern = pattern;
    }

    /**
     * @see javax.jbi.messaging.MessageExchange#setProperty(java.lang.String,
     *      java.lang.Object)
     */
    public void setProperty(final String name, final Object obj) {
        this.properties.put(name, obj);
    }

    public void setRole(final Role role) {
        this.role = role;
    }

    /**
     * set Location ServiceName. A destination Endpoint reference is created
     * with only "serviceName" attribute
     */
    public void setService(final QName service) {
        this.service = service;
    }

    /**
     * @see javax.jbi.messaging.MessageExchange#setStatus(javax.jbi.messaging.
     *      ExchangeStatus)
     */
    public void setStatus(final ExchangeStatus status) throws MessagingException {
        this.checkNotTerminated();

        // check that DONE can be set.
        if (ExchangeStatus.DONE.equals(status)) {
            if (Role.CONSUMER.equals(this.role)) {
                if (IN_ONLY_PATTERN.equals(this.pattern)) {
                    throw new MessagingException(
                    "The MessageExchange state does not allow this operation.");
                } else if (IN_OUT_PATTERN.equals(this.pattern) && this.fault == null
                        && this.getMessage(OUT_MSG) == null) {
                    throw new MessagingException(
                    "The MessageExchange state does not allow this operation.");
                } else if (IN_OPTIONAL_OUT_PATTERN.equals(this.pattern) && this.fault == null
                        && this.getMessage(OUT_MSG) == null) {
                    throw new MessagingException(
                    "The MessageExchange state does not allow this operation.");
                }
            } else {
                if (ROBUST_IN_ONLY_PATTERN.equals(this.pattern) && this.fault != null) {
                    throw new MessagingException(
                    "The MessageExchange state does not allow this operation.");
                } else if (IN_OUT_PATTERN.equals(this.pattern)) {
                    throw new MessagingException(
                    "The MessageExchange state does not allow this operation.");
                } else if (IN_OPTIONAL_OUT_PATTERN.equals(this.pattern)) {
                    if (this.fault != null && this.getMessage(OUT_MSG) == null) {
                        throw new MessagingException(
                        "The MessageExchange state does not allow this operation.");
                    } else if (this.fault == null && this.getMessage(OUT_MSG) != null) {
                        throw new MessagingException(
                        "The MessageExchange state does not allow this operation.");
                    }
                }
            }
        } else if (ExchangeStatus.ERROR.equals(status)) {
            if (Role.PROVIDER.equals(this.role) && IN_OUT_PATTERN.equals(this.pattern)) {
                throw new MessagingException(
                "The MessageExchange state does not allow this operation.");
            }
        }
        this.status = status;
    }

    public void setTerminated(final boolean terminated) {
        this.terminated = terminated;
    }

    /**
     * @param transacted
     *            {@code true} is this exchange is transacted
     * 
     */
    public void setTransacted(final boolean transacted) {
        this.transacted = transacted;
    }

    @Override
    public String toString() {
        return "MessageExchange@" + this.getExchangeId();
    }

    /**
     * Check if the exchange is still valid.
     * 
     * @throws MessagingException
     */
    protected void checkNotTerminated() throws MessagingException {
        if (this.isTerminated()) {
            throw new MessagingException("The Exchange is terminated.");
        }
    }

    /**
     * Check that the exchange MEP allows the specified reference. This only
     * works for the JBI predefined MEPs. It check that reference is IN or OUT.
     * Other reference or MEP will cause an exception
     * 
     * @param name
     * @throws MessagingException
     *             the MEP does not allow the reference
     */
    protected void checkPatternMatching(final String name) throws MessagingException {
        if (this.pattern == null) {
            throw new MessagingException("The MEP is not defined.");
        }

        if (name == null) {
            throw new MessagingException("The reference name is not defined.");
        }

        if (IN_MSG.equalsIgnoreCase(name)) {
            if (IN_ONLY_PATTERN.equals(this.pattern) || ROBUST_IN_ONLY_PATTERN.equals(this.pattern)
                    || IN_OUT_PATTERN.equals(this.pattern)
                    || IN_OPTIONAL_OUT_PATTERN.equals(this.pattern)) {
                return;
            }
        } else if (OUT_MSG.equalsIgnoreCase(name)
                && (IN_OUT_PATTERN.equals(this.pattern) || IN_OPTIONAL_OUT_PATTERN
                        .equals(this.pattern))) {
            return;
        }
        throw new MessagingException("the MessageExchange state does not allow this operation.");
    }

    /**
     * Depending on the state of the Exchange, clean IN and/or OUT normalized
     * messages to reduce exchange size. Status DONE, ERROR : all messages are
     * cleared Fault message : all other messages are cleared OUT message set :
     * IN message is cleared
     */
    public void cleanMessages() {
        // if status is DONE or ERROR, do not send
        // Normalized messages
        if (!ExchangeStatus.ACTIVE.equals(this.status)) {
            this.messages.clear();
            this.fault = null;
        }
        // if there is a fault, no other messages are kept
        if (this.fault != null) {
            this.messages.clear();
        }
        // if OUT response, IN msg is not kept
        if (this.messages.containsKey(OUT_MSG) && this.messages.containsKey(IN_MSG)) {
            this.messages.remove(IN_MSG);
        }
    }

    /**
     * @see MessageExchangeImpl#readObject(ObjectInputStream)
     */
    protected void readObjectDelegate(final ObjectInputStream s) throws IOException {
        // Role deserialization
        switch (s.readShort()) {
            case 0:
                this.role = MessageExchange.Role.CONSUMER;
                break;
            case 1:
                this.role = MessageExchange.Role.PROVIDER;
                break;
            default:
                break;
        }

        // Status deserialization
        switch (s.readShort()) {
            case 0:
                this.status = ExchangeStatus.ACTIVE;
                break;
            case 1:
                this.status = ExchangeStatus.DONE;
                break;
            case 2:
                this.status = ExchangeStatus.ERROR;
                break;
            default:
                break;
        }

        try {
            s.defaultReadObject();
        } catch (final ClassNotFoundException e) {
            throw new IOException(e.getClass() + ":" + e.getMessage());
        }
    }

    /**
     * @see MessageExchangeImpl#writeObject(ObjectOutputStream)
     */
    protected void writeObjectDelegate(final ObjectOutputStream s) throws IOException {
        // Role serialization
        if (MessageExchange.Role.CONSUMER.equals(this.role)) {
            s.writeShort(SERIALIZE_ROLE_CONSUMER);
        } else {
            s.writeShort(SERIALIZE_ROLE_PROVIDER);
        }

        // Status serialization
        if (ExchangeStatus.ACTIVE.equals(this.status)) {
            s.writeShort(SERIALIZE_STATUS_ACTIVE);
        } else if (ExchangeStatus.DONE.equals(this.status)) {
            s.writeShort(SERIALIZE_STATUS_DONE);
        } else {
            s.writeShort(SERIALIZE_STATUS_ERROR);
        }
        s.defaultWriteObject();
    }

    /**
     * Used to deserialize Status and Role objects
     * 
     * @param s
     * @throws IOException
     */
    private void readObject(final ObjectInputStream s) throws IOException {
        this.readObjectDelegate(s);
    }

    /**
     * Used to serialize Status and Role objects
     * 
     * @param s
     * @throws IOException
     */
    private void writeObject(final ObjectOutputStream s) throws IOException {
        this.writeObjectDelegate(s);
    }


    /**
     * {@inheritDoc}
     */
    public void setConsumerEndpoint(final ServiceEndpoint serviceEndpoint) {
        this.consumerEndpoint = serviceEndpoint;
    }

    public void setFlowAttributes(FlowAttributes flowAttributes) {
        this.flowAttributes = flowAttributes;
    }

    public FlowAttributes getFlowAttributes() {
        return this.flowAttributes;
    }

}
